@@ -21,8 +21,10 @@ package org.apache.johnzon.core;
 import javax.json.JsonException;
 import javax.json.stream.JsonLocation;
 import javax.json.stream.JsonParsingException;
+
 import java.io.IOException;
 import java.io.InputStream;
+import java.io.InputStreamReader;
 import java.io.Reader;
 import java.math.BigDecimal;
 import java.nio.charset.Charset;
@@ -40,6 +42,9 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
     //-1 would cause a re-read of the first character in the buffer (which is at zero index)
     private int bufferPos = Integer.MIN_VALUE;
 
+    // performance optimisation to avoid subtraction on readNextChar
+    private int bufferLeft = 0;
+
     //available character in the buffer. It might be <= "buffer.length".
     private int availableCharsInBuffer;
 
@@ -94,6 +99,9 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
     private StructureElement currentStructureElement = null;
 
     private int arrayDepth = 0;
+    private int objectDepth = 0;
+
+    private boolean closed;
 
     //minimal stack implementation
     private static final class StructureElement {
@@ -110,29 +118,30 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
     //detect charset according to RFC 4627
     public JsonStreamParserImpl(final InputStream inputStream, final int maxStringLength,
                                 final BufferStrategy.BufferProvider<char[]> bufferProvider, final BufferStrategy.BufferProvider<char[]> valueBuffer,
-                                final boolean autoAdjust) {
+                                final boolean autoAdjust, final JsonProviderImpl provider) {
 
-        this(inputStream, null, null, maxStringLength, bufferProvider, valueBuffer, autoAdjust);
+        this(inputStream, null, null, maxStringLength, bufferProvider, valueBuffer, autoAdjust, provider);
     }
 
     //use charset provided
     public JsonStreamParserImpl(final InputStream inputStream, final Charset encoding, final int maxStringLength,
                                 final BufferStrategy.BufferProvider<char[]> bufferProvider, final BufferStrategy.BufferProvider<char[]> valueBuffer,
-                                final boolean autoAdjust) {
+                                final boolean autoAdjust, final JsonProviderImpl provider) {
 
-        this(inputStream, null, encoding, maxStringLength, bufferProvider, valueBuffer, autoAdjust);
+        this(inputStream, null, encoding, maxStringLength, bufferProvider, valueBuffer, autoAdjust, provider);
     }
 
     public JsonStreamParserImpl(final Reader reader, final int maxStringLength, final BufferStrategy.BufferProvider<char[]> bufferProvider,
-                                final BufferStrategy.BufferProvider<char[]> valueBuffer, final boolean autoAdjust) {
+                                final BufferStrategy.BufferProvider<char[]> valueBuffer, final boolean autoAdjust, final JsonProviderImpl provider) {
 
-        this(null, reader, null, maxStringLength, bufferProvider, valueBuffer, autoAdjust);
+        this(null, reader, null, maxStringLength, bufferProvider, valueBuffer, autoAdjust, provider);
     }
 
     private JsonStreamParserImpl(final InputStream inputStream, final Reader reader, final Charset encoding, final int maxStringLength,
                                  final BufferStrategy.BufferProvider<char[]> bufferProvider, final BufferStrategy.BufferProvider<char[]> valueBuffer,
-                                 final boolean autoAdjust) {
+                                 final boolean autoAdjust, final JsonProviderImpl provider) {
 
+        super(provider);
         this.autoAdjust = autoAdjust;
         this.maxValueLength = maxStringLength <= 0 ? 8192 : maxStringLength;
         this.fallBackCopyBuffer = valueBuffer.newBuffer();
@@ -146,13 +155,18 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
 
         if (reader != null) {
             this.in = reader;
-        } else {
-            this.in = new RFC4627AwareInputStreamReader(inputStream, encoding);
+        } else if (encoding != null) { // always respect it
+            this.in = new InputStreamReader(inputStream, encoding);
+        } else { // should we log the usage is unexpected? (for perf we want to avoid the pushbackinputstream)
+            this.in = new RFC4627AwareInputStreamReader(inputStream);
         }
     }
 
     //append a single char to the value buffer
     private void appendToCopyBuffer(final char c) {
+        if (fallBackCopyBufferLength >= fallBackCopyBuffer.length - 1) {
+            doAutoAdjust(1);
+        }
         fallBackCopyBuffer[fallBackCopyBufferLength++] = c;
     }
 
@@ -166,19 +180,7 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
             }
 
             if (fallBackCopyBufferLength >= fallBackCopyBuffer.length - length) { // not good at runtime but handled
-                if (!autoAdjust) {
-                    throw new ArrayIndexOutOfBoundsException("Buffer too small for such a long string");
-                }
-
-                final char[] newArray = new char[fallBackCopyBuffer.length + Math.max(getBufferExtends(fallBackCopyBuffer.length), length)];
-                // TODO: log to adjust size once?
-                System.arraycopy(fallBackCopyBuffer, 0, newArray, 0, fallBackCopyBufferLength);
-                System.arraycopy(buffer, startOfValueInBuffer, newArray, fallBackCopyBufferLength, length);
-                if (releaseFallBackCopyBufferLength) {
-                    bufferProvider.release(fallBackCopyBuffer);
-                    releaseFallBackCopyBufferLength = false;
-                }
-                fallBackCopyBuffer = newArray;
+                doAutoAdjust(length);
             } else {
                 System.arraycopy(buffer, startOfValueInBuffer, fallBackCopyBuffer, fallBackCopyBufferLength, length);
             }
@@ -188,7 +190,26 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
         startOfValueInBuffer = endOfValueInBuffer = -1;
     }
 
+    private void doAutoAdjust(final int length) {
+        if (!autoAdjust) {
+            throw new ArrayIndexOutOfBoundsException("Buffer too small for such a long string");
+        }
+
+        final char[] newArray = new char[fallBackCopyBuffer.length + Math.max(getBufferExtends(fallBackCopyBuffer.length), length)];
+        // TODO: log to adjust size once?
+        System.arraycopy(fallBackCopyBuffer, 0, newArray, 0, fallBackCopyBufferLength);
+        if (startOfValueInBuffer != -1) {
+            System.arraycopy(buffer, startOfValueInBuffer, newArray, fallBackCopyBufferLength, length);
+        }
+        if (releaseFallBackCopyBufferLength) {
+            bufferProvider.release(fallBackCopyBuffer);
+            releaseFallBackCopyBufferLength = false;
+        }
+        fallBackCopyBuffer = newArray;
+    }
+
     /**
+     * @param currentLength length of the buffer
      * @return the amount of bytes the current buffer should get extended with
      */
     protected int getBufferExtends(int currentLength) {
@@ -198,13 +219,17 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
 
     @Override
     public final boolean hasNext() {
-
-        if (currentStructureElement != null ||
-                (previousEvent != END_ARRAY && previousEvent != END_OBJECT &&
-                        previousEvent != VALUE_STRING && previousEvent != VALUE_FALSE && previousEvent != VALUE_TRUE &&
-                        previousEvent != VALUE_NULL && previousEvent != VALUE_NUMBER) ||
-                previousEvent == 0) {
-
+        if (currentStructureElement != null || previousEvent == 0) {
+            return true;
+        }
+        if (previousEvent != END_ARRAY && previousEvent != END_OBJECT &&
+                previousEvent != VALUE_STRING && previousEvent != VALUE_FALSE && previousEvent != VALUE_TRUE &&
+                previousEvent != VALUE_NULL && previousEvent != VALUE_NUMBER) {
+            if (bufferPos < 0) { // check we don't have an empty string to parse
+                final char c = readNextChar();
+                unreadChar();
+                return c != EOF;
+            }
             return true;
         }
 
@@ -271,7 +296,7 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
     //refill is necessary copy the already read value part into the value buffer
     protected final char readNextChar() {
 
-        if ((availableCharsInBuffer - bufferPos) <= 1) {
+        if (bufferLeft == 0) {
             //fillbuffer
 
             //copy content from old buffer to valuebuffer
@@ -299,6 +324,7 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
             }
 
             bufferPos = 0;
+            bufferLeft = availableCharsInBuffer - 1;
             //end fillbuffer
         } else {
 
@@ -309,6 +335,7 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
             //}
 
             bufferPos++;
+            bufferLeft--;
         }
 
         return buffer[bufferPos];
@@ -345,18 +372,28 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
     @Override
     public Event current() {
         if (previousEvent < 0 && hasNext()) {
-            next();
+            internalNext();
         }
         return previousEvent >= 0 && previousEvent < Event.values().length
                 ? Event.values()[previousEvent]
                 : null;
     }
 
+    private void unreadChar() {
+        bufferPos--;
+        bufferLeft++;
+    }
+
     @Override
-    public final Event next() {
+    protected final Event internalNext() {
         //main entry, make decision how to handle the current character in the stream
 
         if (!hasNext()) {
+            final char c = readNextChar();
+            unreadChar();
+            if (c != EOF) {
+                throw uexc("No available event");
+            }
             throw new NoSuchElementException();
         }
 
@@ -375,7 +412,7 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
             }
 
             previousEvent = COMMA_EVENT;
-            return next();
+            return internalNext();
 
         }
 
@@ -386,7 +423,7 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
             }
 
             previousEvent = KEY_SEPARATOR_EVENT;
-            return next();
+            return internalNext();
 
         }
 
@@ -477,6 +514,8 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
             currentStructureElement = localStructureElement;
         }
 
+        objectDepth++;
+
         return EVT_MAP[previousEvent = START_OBJECT];
 
     }
@@ -496,6 +535,8 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
         //pop from stack
         currentStructureElement = currentStructureElement.previous;
 
+        objectDepth--;
+
         return EVT_MAP[previousEvent = END_OBJECT];
     }
 
@@ -547,6 +588,17 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
         return arrayDepth > 0;
     }
 
+
+    @Override
+    protected boolean isInObject() {
+        return objectDepth > 0;
+    }
+
+    @Override
+    protected BufferStrategy.BufferProvider<char[]> getCharArrayProvider() {
+        return bufferProvider;
+    }
+
     //read a string, gets called recursively
     //Handles escape/d characters
     //if string contains escape chars and/or cross buffer boundary then copy in the value buffer
@@ -583,7 +635,6 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
                     //another escape chars, for example \t
                 } else {
                     appendToCopyBuffer(Strings.asEscapedChar(n));
-
                 }
 
             } else {
@@ -620,7 +671,7 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
 
                 //current n is one of < '\u001F' -OR- ESCAPE_CHAR -OR- EOL -OR- QUOTE
 
-                bufferPos--; //unread one char
+                unreadChar(); //unread one char
 
             }
         } while (true);
@@ -635,33 +686,6 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
 
     }
 
-    //maybe we want to check invalid utf encoding
-    //not clear yet if the InputStreamReader is doing that
-
-    /*
-    private char checkSurrogates(char n, char highSurrogate) {
-        //check for invalid surrogates
-        //high followed by low       
-        if (Character.isHighSurrogate(n)) {
-
-            if (highSurrogate != 0) {
-                throw uexc("Unexpected high surrogate");
-            }
-            return n;
-        } else if (Character.isLowSurrogate(n)) {
-
-            if (highSurrogate == 0) {
-                throw uexc("Unexpected low surrogate");
-            } else if (!Character.isSurrogatePair(highSurrogate, n)) {
-                throw uexc("Invalid surrogate pair");
-            }
-            return 0;
-        } else if (highSurrogate != 0 && !Character.isLowSurrogate(n)) {
-            throw uexc("Expected low surrogate");
-        }
-        
-        return highSurrogate;
-    }*/
 
     //read the next four chars, check them and treat them as an single unicode char
     private char parseUnicodeHexChars() {
@@ -676,8 +700,11 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
         //always the beginning quote of a key or value  
 
         //last event must one of the following-> : { [ ,
-        if (previousEvent != -1 && (previousEvent != KEY_SEPARATOR_EVENT && previousEvent != START_OBJECT && previousEvent != START_ARRAY
-                && previousEvent != COMMA_EVENT)) {
+        if (previousEvent != -1 &&
+                (previousEvent != KEY_SEPARATOR_EVENT &&
+                 previousEvent != START_OBJECT &&
+                 previousEvent != START_ARRAY  &&
+                 previousEvent != COMMA_EVENT)) {
             throw uexc("Expected : { [ ,");
         }
         //starting quote already consumed
@@ -782,7 +809,7 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
 
         if (y == COMMA_CHAR || y == END_ARRAY_CHAR || y == END_OBJECT_CHAR || y == EOL || y == SPACE || y == TAB || y == CR || y == EOF) {
 
-            bufferPos--;//unread one char
+            unreadChar();//unread one char
 
             //['-', DIGIT]
             if (isCurrentNumberIntegral && c == MINUS && cumulatedDigitValue >= 48 && cumulatedDigitValue <= 57) {
@@ -930,6 +957,18 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
 
     }
 
+    @Override
+    public boolean isFitLong() { // not exact but good enough for most cases
+        if (!isCurrentNumberIntegral) {
+            return false;
+        }
+
+        // no buffer overflow - assumes a buffer can hold a long
+        // + length <= since max long is 9223372036854775807 and min is -9223372036854775808
+        final int len = endOfValueInBuffer - startOfValueInBuffer;
+        return fallBackCopyBufferLength <= 0 && len > 0 && len <= 18;
+    }
+
     @Override
     public BigDecimal getBigDecimal() {
         if (previousEvent != VALUE_NUMBER) {
@@ -938,23 +977,10 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
             //            return currentBigDecimalNumber;
         } else if (isCurrentNumberIntegral && currentIntegralNumber != Integer.MIN_VALUE) {
             return new BigDecimal(currentIntegralNumber);
-        } else if (isCurrentNumberIntegral) {
-            //if there a content in the value buffer read from them, if not use main buffer
-            final Long retVal = fallBackCopyBufferLength > 0 ? parseLongFromChars(fallBackCopyBuffer, 0, fallBackCopyBufferLength)
-                    : parseLongFromChars(buffer, startOfValueInBuffer, endOfValueInBuffer);
-            if (retVal == null) {
-                return (/*currentBigDecimalNumber = */fallBackCopyBufferLength > 0 ? new BigDecimal(fallBackCopyBuffer, 0,
-                        fallBackCopyBufferLength) : new BigDecimal(buffer, startOfValueInBuffer,
-                        (endOfValueInBuffer - startOfValueInBuffer)));
-            } else {
-                return (/*currentBigDecimalNumber = */new BigDecimal(retVal.longValue()));
-            }
-        } else {
-            //if there a content in the value buffer read from them, if not use main buffer
-            return (/*currentBigDecimalNumber = */fallBackCopyBufferLength > 0 ? new BigDecimal(fallBackCopyBuffer, 0,
-                    fallBackCopyBufferLength) : new BigDecimal(buffer, startOfValueInBuffer, (endOfValueInBuffer - startOfValueInBuffer)));
         }
-
+        //if there a content in the value buffer read from them, if not use main buffer
+        return (/*currentBigDecimalNumber = */fallBackCopyBufferLength > 0 ? new BigDecimal(fallBackCopyBuffer, 0,
+                fallBackCopyBufferLength) : new BigDecimal(buffer, startOfValueInBuffer, (endOfValueInBuffer - startOfValueInBuffer)));
     }
 
     @Override
@@ -964,6 +990,10 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
 
     @Override
     public void close() {
+        if (closed) {
+            return;
+        }
+
         bufferProvider.release(buffer);
         if (releaseFallBackCopyBufferLength) {
             valueProvider.release(fallBackCopyBuffer);
@@ -973,6 +1003,8 @@ public class JsonStreamParserImpl extends JohnzonJsonParserImpl implements JsonC
             in.close();
         } catch (final IOException e) {
             throw new JsonException("Unexpected IO exception " + e.getMessage(), e);
+        } finally {
+            closed = true;
         }
     }
 
